#include "indexator.h"
/** il faut convertir les titres en minuscules **/

/* {{{ structures */
/* contient les infos sur le fichier contenant les recettes */
struct Trecette_info {    
    unsigned long offset_titre;	/* rindex.pinfo[4].off_t = titre de la 4 */
    unsigned long strlen_titre;	/* longueur de la chaine, sans le \0 final */
    unsigned long offset_corps;
    unsigned long strlen_corps;
};
struct Tfichier_index {
    unsigned long nb_recettes;	/* nombre de recettes de cette structure (0..n-1) */
    struct Trecette_info * pinfotab;
};

/* contient toutes les recettes lues dans le fichier */
struct Tchartab {
    char * str;
};
struct Tstruct_recettes {
    unsigned long nb_recettes;
    struct Tchartab * ptitretab; /* recette.ptitretab[3].str = (char *) titre de la 4 recette */
    struct Tchartab * pcorpstab;
};
/* }}} */
struct Tstruct_recettes ensemblerecettes = {0, NULL, NULL};
/* {{{ recette_titre(), recette_corps(), nbrecettes() */
char * recette_titre (unsigned long recettenum) {
    if ((recettenum >= ensemblerecettes.nb_recettes) || (recettenum < 0))
	return NULL;
    return ensemblerecettes.ptitretab[recettenum].str;
}

char * recette_corps (unsigned long recettenum) {
    if ((recettenum >= ensemblerecettes.nb_recettes) || (recettenum < 0))
	return NULL;
    return ensemblerecettes.pcorpstab[recettenum].str;
}

unsigned long nbrecettes () {
    return ensemblerecettes.nb_recettes;
}
/* }}} */
/* {{{ create_index() */
struct Tfichier_index create_index (char * dbfname) {
    struct Tfichier_index rindex = {0, NULL};
    FILE * DB = NULL;
    char c;		/* utilis pour la lecture du fichier caracteres par caracteres */
    register long n = -1;
    int i = 0;

    if ((DB = fopen (dbfname, "r")) == NULL) {
	strcpy (errorstr, "createindex : impossible d'ouvrir le fichier de recettes");
	return rindex;
    }
    
    while (fread (&c, sizeof (char), 1, DB) > 0) {
    	/* on ne s'intresse qu'a '$' */
    	
/* exemple de recette telle que dans le fichier :
$ Titre de la recette ici $@CODE@@ALIMENTS@

La recette est compose de son corps

constitu de plusieurs lignes

et termine comme ca :$

-------------
$ Titre suivant ...
*/
	if (c != '$')
	    continue;
	
	/* 3 $ par recette */
	i++; /* i dans {1, 2, 3} */
	switch (i) {
	case 1: /* au debut du titre */
	    n++;
	    
/* on alloue de la place pour une nouvelle structure pour contenir une recette supplmentaire */
	    rindex.pinfotab = (struct Trecette_info *)realloc (rindex.pinfotab, (n + 1) * sizeof (struct Trecette_info));

/* on connait dja l'offset du dbut du titre : */
	    rindex.pinfotab[n].offset_titre = ftell (DB) + 1;
	    break;
	    
	case 2: /* fin titre */
/* on calcule l'offset de fin du titre et la longueur du titre */
	    rindex.pinfotab[n].strlen_titre = ftell (DB) - rindex.pinfotab[n].offset_titre - 1;
	    
/* la ligne est ensuite utilise par des infos que l'on ignore */
	    while (fread (&c, sizeof (char), 1, DB) > 0)
		if (c == '\n')
		    break;
/* on est alors au dbut du corps */
	    rindex.pinfotab[n].offset_corps = ftell (DB) + 1;
	    break;
	    
	case 3: /* fin corps */
	    rindex.pinfotab[n].strlen_corps = ftell (DB) - rindex.pinfotab[n].offset_corps - 1;
	    i = 0;
	}
    }
    if (feof (DB)) {
/* on est arriv normalement a la fin du fichier */
	rindex.nb_recettes = n + 1;
	return rindex;
    } /* erreur */
    free (rindex.pinfotab);
    rindex.pinfotab = NULL;
    rindex.nb_recettes = 0;
    strcpy (errorstr, "createindex : lecture des recettes");
    return rindex;
}
/* }}} */
/* {{{ save_index() et load_index() */
/* essaye de sauver l'index sur disque */
void save_index (char * indexfname, struct Tfichier_index rindex) {
    FILE * fdindex = NULL;
    unsigned long l;
    
    if ((fdindex = fopen (indexfname, "w")) == NULL) {
/* l'erreur n'est pas fatale : il suffit de recrr l'index a chaque fois */
	return;
    }

/* on crit le nombre de recette qu'est cens contenir le fichier */   
    fwrite (&rindex.nb_recettes, sizeof (unsigned long), 1, fdindex);
    
/* on crit l'ensemble des rfrences */
    l = fwrite (rindex.pinfotab, sizeof (struct Trecette_info), rindex.nb_recettes, fdindex);
    
/* debug */
    printf ("debug : %ld references de recettes ecrites dans l'index %s\n", l, indexfname);
    return;
}

struct Tfichier_index load_index (char * indexfname) {
    struct Tfichier_index rindex = {0, NULL};
    FILE * fdindex = NULL;
    unsigned long l;
    
    if ((fdindex = fopen (indexfname, "r")) == NULL) {
	strcpy (errorstr, "createindex : ne peut ouvrir le fichier d'index en lecture");
/* dans ce cas il faudra refaire l'index sans le sauvegarder */
	return rindex;
    }

/* on lit le nombre de recettes de l'index */
    fread (&rindex.nb_recettes, sizeof (unsigned long), 1, fdindex);

    rindex.pinfotab = (struct Trecette_info *)malloc ((rindex.nb_recettes) * sizeof (struct Trecette_info));
    if (rindex.pinfotab == NULL) {
	rindex.nb_recettes = 0;
	strcpy (errorstr, "loadindex : malloc ()");
	return rindex;
    }
    
/* on lit l'ensemble de donnes sur le fichier de recettes d'un coup */
    l = fread (rindex.pinfotab, sizeof (struct Trecette_info), rindex.nb_recettes, fdindex);
    printf ("debug : %ld references lues depuis %s\n", l, indexfname);
    return rindex;
}
/* }}} */
/* {{{ load_recettes() et destroy_recettes() */
void str2lower (char * str) {
    while (*str != '\0') {
    	if ((*str >= 'A') && (*str <= 'Z'))
	    *str += 'a' - 'A';
	str++;
    }
}

void load_recettes (char * dbfname, char * indexfname) {
    struct Tfichier_index rindex = {0, NULL};
    FILE * DB;
    unsigned long i, l;
    
    printf ("debug : debut de loadrecettes (%s, %s)\n", dbfname, indexfname);
    if (dbfname == NULL)
	return;
	
/* on regarde si l'index existe dja */
    if (indexfname != NULL) {
        if (access (indexfname, F_OK)) {
/* si il n'existe pas, on le cre et on le sauvegarde */
    printf ("debug : l'index n'existe pas, dbut de cration\n");
	    rindex = create_index (dbfname);
    printf ("debug : cration termine\n");
	    if (rindex.nb_recettes > 0) {
    printf ("debug : sauvegarde de l'index\n");
		save_index (indexfname, rindex);
    printf ("debug : sauvegarde terminee\n");
	    }
	} else {
    printf ("debug : l'index existe, dbut de chargement\n");
	    rindex = load_index (indexfname);
    printf ("debug : chargement termin\n");
	    if (rindex.nb_recettes <= 0) {
    printf ("debug : 0 recettes chargees : recration\n");
		rindex = create_index (dbfname);
    printf ("debug : recreation termine\n");
		unlink (indexfname);
    printf ("debug : sauvegarde de l'index\n");
		save_index (indexfname, rindex);
    printf ("debug : sauvegarde terminee\n");
	    }
	}
    } else {
    printf ("debug : pas de nom d'index donn, cration\n");
    	rindex = create_index (dbfname);
    printf ("debug : creation termine\n");
    }

    if ((DB = fopen (dbfname, "r")) == NULL) {
    	free (rindex.pinfotab);
	strcpy (errorstr, "loadrecettes : impossible de lire le fichier de recetes");
/* on ne fait rien de particulier : si c'est la premiere tentative, l'appelant verra nbrecette () == 0,
 * et agira en consquences, sinon on reste avec le meme ensemble de recettes */
	return;
    }

/* on libere la structure au cas ou ce ne serait pas le premier appel */
    destroy_recettes ();

    printf ("debug : dbut de lecture des recettes\n");

/* on alloue l'espace pour les tableaux de char * */
    ensemblerecettes.ptitretab = (struct Tchartab *)malloc ((rindex.nb_recettes) * sizeof (struct Tchartab));
    ensemblerecettes.pcorpstab = (struct Tchartab *)malloc ((rindex.nb_recettes) * sizeof (struct Tchartab));
    if ((ensemblerecettes.ptitretab == NULL) || (ensemblerecettes.ptitretab == NULL)) {
	if (ensemblerecettes.ptitretab != NULL) {
	    free (ensemblerecettes.ptitretab);
	    ensemblerecettes.ptitretab = NULL;
	}
	if (ensemblerecettes.pcorpstab != NULL) {
	    free (ensemblerecettes.pcorpstab);
	    ensemblerecettes.pcorpstab = NULL;
	}
	strcpy (errorstr, "loadrecettes : malloc ()");
	free (rindex.pinfotab);
	return;
    }

    ensemblerecettes.nb_recettes = rindex.nb_recettes;
    for (i = 0; i < rindex.nb_recettes; i++) {
/* on lit le titre a la position indique dans l'index */
	l = rindex.pinfotab[i].strlen_titre + 1;
	if ((ensemblerecettes.ptitretab[i].str = (char *)malloc (l)) == NULL) {
/* si on ne peut allouer pour une recette, on quitte dans cet tat */
	    ensemblerecettes.nb_recettes = i - 1;
	    strcpy (errorstr, "loadrecettes : malloc ()");
	    break;
	}
/* on initialise l'espace allou a la chaine a 0 de maniere a obtenir a la fin de la chaine un \0 */
	bzero (ensemblerecettes.ptitretab[i].str, l--);
/* on se positionne suivant l'index et on lit le bon nombre de caracteres */
	fseek (DB, rindex.pinfotab[i].offset_titre, SEEK_SET);
	fread (ensemblerecettes.ptitretab[i].str, sizeof (char), l, DB);
	str2lower (ensemblerecettes.ptitretab[i].str);

	l = rindex.pinfotab[i].strlen_corps + 1;
	if ((ensemblerecettes.pcorpstab[i].str = (char *)malloc (l)) == NULL) {
	    free (ensemblerecettes.ptitretab[i].str);
	    ensemblerecettes.nb_recettes = i - 1;
	    strcpy (errorstr, "loadrecettes : malloc ()");
	    break;
	}
	bzero (ensemblerecettes.pcorpstab[i].str, l--);
	fseek (DB, rindex.pinfotab[i].offset_corps, SEEK_SET);
	fread (ensemblerecettes.pcorpstab[i].str, sizeof (char), l, DB);
    }
    free (rindex.pinfotab);
}

void destroy_recettes () {
    unsigned long i;
    if ((ensemblerecettes.nb_recettes = 0))
	return;
    for (i = 0; i < ensemblerecettes.nb_recettes; i++) {
	free (ensemblerecettes.ptitretab[i].str);
	free (ensemblerecettes.pcorpstab[i].str);
    }
    free (ensemblerecettes.ptitretab);
    ensemblerecettes.ptitretab = NULL;
    free (ensemblerecettes.pcorpstab);
    ensemblerecettes.pcorpstab = NULL;
    ensemblerecettes.nb_recettes = 0;
}
/* }}} */
/*@}*/
 
